import copy
import math
from enum import Enum, auto
from typing import Dict, List, Optional

from pydantic import BaseModel, Field, RootModel, constr

Date = constr(pattern=r"\d{2}/\d{2}/\d{4}")


class LevelsInfoEnum(Enum):
    LEVEL_1 = "Level1"
    LEVEL_2 = "Level2"


class PromotionType(Enum):
    URL_REWARD = "urlreward"  # Activity completed by visiting the URL
    QUIZ = "quiz"  # Answer from 5 to 10 questions

    SEARCH = "search"  # PC Search
    EMPTY = ""  # prevents enum casting to fail in case of empty string

    WELCOME_TOUR = "welcometour"  # Same as url reward, but granted only once

    # streak keepers
    STREAK = "streak"
    STREAK_BONUS = "streakbonus"
    COACH_MARKS = "coachmarks"


class QuizType(Enum):
    """
    Differentiation of possible Quiz types.
    """

    # This quiz is a simple choice between two elements.
    # It does not have a Start button to click.
    CHOICE_BETWEEN_TWO = auto()

    # These quizzes have a Start button to click.
    # They can be either questions with four or eight answers.
    THREE_QUESTIONS_FOUR_ANSWERS = auto()
    THREE_QUESTIONS_EIGHT_ANSWERS = auto()


class Promotion(BaseModel):
    name: str
    offerId: str
    complete: bool
    counter: int
    activityProgress: int
    activityProgressMax: int
    pointProgressMax: int
    pointProgress: int
    promotionType: PromotionType
    promotionSubtype: str
    title: str
    extBannerTitle: str
    titleStyle: str
    theme: str
    description: str
    showcaseTitle: str
    showcaseDescription: str
    imageUrl: str
    dynamicImage: str
    destinationUrl: str
    linkText: str
    hash: str
    activityType: str
    isRecurring: bool
    isHidden: bool
    isTestOnly: bool
    isGiveEligible: bool

    def is_enabled(self) -> bool:
        return not self.isHidden and not self.isTestOnly

    def is_completable(self) -> bool:
        return (
            self.is_enabled()
            and not self.complete
            and self.promotionType
            in (
                PromotionType.URL_REWARD,
                PromotionType.QUIZ,
                PromotionType.WELCOME_TOUR,
            )
        )


class LevelInfo(BaseModel):
    activeLevel: LevelsInfoEnum


class SearchCounter(BaseModel):
    name: str
    offerId: str
    complete: bool
    pointProgress: int
    pointProgressMax: int
    description: str

    def is_completable(self) -> bool:
        return not self.complete

    @staticmethod
    def get_points_per_search() -> int:
        """
        Returns the amount of points earned per single search.
        """

        return 3  # TODO: this could be also parsed from `description`

    def get_needed_searches_amount(self) -> int:
        """
        Returns the amount of searches needed to reach the maximum for the day.
        """

        delta_points = self.pointProgressMax - self.pointProgress
        count_searches: float = delta_points / self.get_points_per_search()
        count_searches_ceil = math.ceil(count_searches)
        return int(count_searches_ceil)


class Counters(BaseModel):
    # two items expected: actual pc searches counter, and bing searches counter
    pcSearch: List[SearchCounter] = Field(..., min_length=2, max_length=2)

    # one item expected: actual mobile searches counter
    mobileSearch: Optional[List[SearchCounter]] = Field(
        None, min_length=1, max_length=1
    )


class UserStatus(BaseModel):
    levelInfo: LevelInfo
    availablePoints: int
    counters: Counters


# RootModel instead of BaseModel because of:
# https://docs.pydantic.dev/latest/usage/models/#rootmodel-and-custom-root-types
class DailySetPromotions(RootModel):
    # `root` is needed to parse automatically `Date` keys, since they are dynamic
    # Minimum 2 items (daily set for today and tomorrow),
    # but occasionally 3 items are available (yesterday, today, tomorrow)
    root: Dict[Date, List[Promotion]] = Field(min_length=2, max_length=3)  # type: ignore

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def _keys(self) -> List[str]:
        """
        Returns the keys found for storing the DailySets.
        """

        return list(self.root.keys())

    def get_daily_set_for_today(self) -> List[Promotion]:
        """
        Returns the daily set that is available today.
        """

        keys = self._keys()
        if len(keys) == 2:
            key = keys[0]
        elif len(keys) == 3:
            key = keys[1]
        else:
            # Pydantic will ensure that we don't have <2 or >3 keys in `root` dict
            raise RuntimeError("Unreachable")

        return self.root[key]

    def get_daily_set_for_yesterday(self) -> Optional[List[Promotion]]:
        """
        Returns the daily set that was available yesterday.

        Returns None if this set is not found.
        """

        keys = self._keys()
        if len(keys) == 2:
            return None
        elif len(keys) == 3:
            key = keys[0]
        else:
            # Pydantic will ensure that we don't have <2 or >3 keys in `root` dict
            raise RuntimeError("Unreachable")

        return self.root[key]

    def get_daily_set_for_tomorrow(self) -> List[Promotion]:
        """
        Returns the daily set that will be available tomorrow.
        """

        keys = self._keys()
        if len(keys) == 2:
            key = keys[1]
        elif len(keys) == 3:
            key = keys[2]
        else:
            # Pydantic will ensure that we don't have <2 or >3 keys in `root` dict
            raise RuntimeError("Unreachable")

        return self.root[key]


class Dashboard(BaseModel):
    userStatus: UserStatus
    promotionalItem: Optional[Promotion]
    dailySetPromotions: DailySetPromotions
    # punchCards: List[Any]
    morePromotions: List[Promotion]

    def level(self) -> LevelsInfoEnum:
        """
        Returns the current user level.
        """

        return self.userStatus.levelInfo.activeLevel

    def points(self) -> int:
        """
        Returns the amount of points available.
        """

        return self.userStatus.availablePoints

    def can_search_on_pc(self) -> bool:
        """
        Returns True if any PC search is missing, False otherwise.
        """

        return self.userStatus.counters.pcSearch[0].is_completable()

    def can_search_on_mobile(self) -> bool:
        """
        Returns True if any Mobile search is missing, False otherwise.
        """

        if self.level() is LevelsInfoEnum.LEVEL_1:
            return False

        assert self.userStatus.counters.mobileSearch is not None
        return self.userStatus.counters.mobileSearch[0].is_completable()

    def amount_of_pc_searches(self) -> int:
        """
        Returns the amount of PC searches that are needed to be executed.

        Returns 0 if no PC search is needed, or the corresponding counter is disabled.
        """

        if not self.can_search_on_pc():
            return 0

        return self.userStatus.counters.pcSearch[0].get_needed_searches_amount()

    def amount_of_mobile_searches(self) -> int:
        """
        Returns the amount of Mobile searches that are needed to be executed.

        Returns 0 if no Mobile search is needed, or the corresponding counter is disabled.
        """

        if not self.can_search_on_mobile():
            return 0

        mobile_searches = self.userStatus.counters.mobileSearch
        assert mobile_searches is not None
        mobile_search = mobile_searches[0]
        return mobile_search.get_needed_searches_amount()

    def get_promotions(self) -> List[Promotion]:
        """
        Returns the list of all promotions available in the current dashboard.
        """

        promotions: List[Promotion] = []

        # First, add the daily set promotions,
        # as they have more priority being part of the streak
        today_daily_set = self.dailySetPromotions.get_daily_set_for_today()
        promotions.extend(today_daily_set)

        # Consideration about the daily sets:
        # - Today's is always available to parse and complete.
        # - Tomorrow's is always available to parse, but not to complete.
        # - Yesterday's could be available to parse, but since it doesn't
        #   show up in the Rewards homepage, it is better to not include it
        #   in the parsed promotions.

        # Then add the extra promotions
        promotions.extend(self.morePromotions)

        # Then returns a copy of the list
        return copy.deepcopy(promotions)

    def get_promotional_item(self) -> Optional[Promotion]:
        return self.promotionalItem

    def get_completable_promotions(self) -> List[Promotion]:
        """
        Get the list of all completable promotions in the current dashboard.
        """

        promotions = self.get_promotions()
        completable_promotions = [
            promotion for promotion in promotions if promotion.is_completable()
        ]
        return completable_promotions



import logging
import random
import string
from abc import ABC
from typing import Generator
from urllib.parse import urlparse

from selenium.webdriver.common.keys import Keys

logger = logging.getLogger(__name__)


def is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
    except (AttributeError, ValueError):
        return False
    else:
        return all(field for field in (result.scheme, result.netloc, result.path))


class SearchGenerator(ABC):
    """
    Interface for search generators to be used as generators of
    query for Bing searches
    """

    def query_gen(self) -> Generator[str, None, None]:
        """
        Returns a generator of queries to be used with Bing searches
        """

        raise NotImplementedError

    def sleep_time(self) -> float:
        """
        Returns the time to sleep in seconds between Bing searches
        """

        raise NotImplementedError


class RandomSearchGenerator(SearchGenerator):
    def sleep_time(self):
        return 1.5

    def query_gen(self):
        alphabet = string.ascii_lowercase
        length = 70
        word = "".join(random.choices(alphabet, k=length))
        logger.debug(f"Generated a word of {length} characters: {word}")

        # the first element to return is the word
        logger.debug("Yielding word...")
        yield word

        # then is always returned backspace
        while True:
            logger.debug("Yielding backspace...")
            yield Keys.BACKSPACE



import json
from pathlib import Path
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

import yaml
from attr import define
from email_validator import (
    ValidatedEmail as _ValidatedEmail,
)
from email_validator import (
    validate_email as _validate_email,
)
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, SecretStr
from typing_extensions import Annotated


def validate_version(value: str) -> str:
    if value != "v1":
        raise ValueError("Version unsupported")
    return value


def validate_url(value: str) -> str:
    result = urlparse(value)
    assert all(field for field in (result.scheme, result.netloc, result.path))
    return value


def validate_email(value: Optional[str]) -> Optional[str]:
    if value is None:
        return None
    validated_email: _ValidatedEmail = _validate_email(value)
    return validated_email.normalized


ValidatedVersion = Annotated[str, AfterValidator(validate_version)]
ValidatedURL = Annotated[str, AfterValidator(validate_url)]
ValidatedEmail = Annotated[str, AfterValidator(validate_email)]


@define(frozen=True)
class Defaults:
    desktop_useragent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/115.0.0.0 Safari/537.36 Edg/115.0.1901.188"
    )
    mobile_useragent = (
        "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/115.0.0.0 Mobile Safari/537.3"
    )

    rewards_homepage = "https://rewards.bing.com/"
    bing_homepage = "https://www.bing.com/?scope=web"


class Profile(BaseModel):
    email: ValidatedEmail
    profile: str
    skip: bool = False


class AutomsrConfig(BaseModel):
    """
    >>> profiles = [{"email": "1@gmail.com", "profile": "p1"}, {"email": "2@gmail.com", "profile": "p2"}]
    >>> _ = AutomsrConfig(profiles=profiles)
    """

    profiles: List[Profile] = Field(..., min_length=1)

    rewards_homepage: ValidatedURL = Defaults.rewards_homepage
    bing_homepage: ValidatedURL = Defaults.bing_homepage

    desktop_useragent: str = Defaults.desktop_useragent
    mobile_useragent: str = Defaults.mobile_useragent


class EmailConfig(BaseModel):
    enable: bool = False
    recipient: Optional[ValidatedEmail] = None
    sender: Optional[ValidatedEmail] = None
    sender_password: Optional[SecretStr] = None
    host: Optional[str] = None
    port: Optional[int] = None
    tls: Optional[bool] = False


class SeleniumConfig(BaseModel):
    profiles_root: Path
    chromedriver_path: Path


class Config(BaseModel):
    model_config = ConfigDict(extra="forbid")

    version: ValidatedVersion = "v1"
    automsr: AutomsrConfig
    email: EmailConfig
    selenium: SeleniumConfig

    @classmethod
    def from_yaml(cls, path: Path) -> "Config":
        """
        Load config from a yaml filepath.
        """

        data = yaml.safe_load(open(path))
        return cls.from_dict(data)

    @classmethod
    def from_json(cls, path: Path) -> "Config":
        """
        Load config from a json filepath.
        """

        data = json.load(open(path))
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Config":
        """
        Load config from an in-memory mapping.

        >>> profiles = [{"email": "1@gmail.com", "profile": "p1"}]
        >>> _data = {
        ...     "automsr": {"profiles": profiles},
        ...     "email": {},
        ...     "selenium": {"profiles_root": Path("profiles"), "chromedriver_path": Path("chromedriver.exe")},
        ... }
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Config(version='v1', automsr=..., email=..., selenium=...)
        >>> _data["email"]["sender"] = "invalid-email"
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        pydantic_core._pydantic_core.ValidationError: 1 validation error for Config
        email.sender
          Value error, The email address is not valid...
        ...
        >>> del _data["email"]["sender"]
        >>> _data["version"] = "v2"
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        pydantic_core._pydantic_core.ValidationError: 1 validation error for Config
        version
          Value error, Version unsupported [type=value_error, input_value='v2', input_type=str]
        ...
        """

        return cls(**data)


if __name__ == "__main__":
    import doctest

    failures, _all_tests = doctest.testmod()
    assert not failures



import logging
import time
from typing import Any, Dict, List, Optional

from attr import define, field
from selenium.common import NoSuchElementException
from selenium.webdriver import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from tqdm import tqdm

from automsr.browser.browser import Browser
from automsr.config import Config, Profile
from automsr.datatypes.dashboard import (
    Dashboard,
    Promotion,
    PromotionType,
    QuizType,
)
from automsr.search import RandomSearchGenerator

logger = logging.getLogger(__name__)


class ExecutorException(Exception):
    """
    Base class for Executor exceptions.
    """


class QuizException(ExecutorException):
    """
    Base class for Quiz Promotions exceptions.
    """


class CannotFindStartButtonException(QuizException):
    """
    Exception raised when the Start button for quizzes is expected but not found.
    """


class CannotDetermineQuizTypeException(QuizException):
    """
    Exception raised when the Executor is unable to determine
    the quiz-type of a promotion.
    """


@define
class SingleTargetExecutor:
    """
    Executor class driving the completion of all the tasks associated with Rewards.

    This executor will target a single profile.
    """

    config: Config
    profile: Profile
    browser: Browser = field(init=False)

    def execute(self) -> None:
        """
        Execute the following steps:
        - Open a new browser session with Selenium and a Chrome driver.
        - Retrieval of dashboard json from Rewards page
        - Execution of all completable punchcards
        - Execution of all completable promotions
        - Execution of searches:
            - PC searches (desktop user agent)
            - Mobile searches (mobile user agent)
        """

        # Start a new session
        self.start_session()

        # Retrieve the current dashboard
        dashboard = self.get_dashboard()

        # TODO Execute punchcards

        # Execute promotions
        self.execute_promotions(dashboard=dashboard)

        # Execute both PC and Mobile searches, if needed
        self.execute_pc_searches(dashboard=dashboard)
        self.execute_mobile_searches(dashboard=dashboard)

        # End session
        self.end_session()

    def start_session(self) -> None:
        """
        Create a new session with Selenium and Chromedriver,
        then returns the session object to the caller.
        """

        self.browser = Browser.from_config(config=self.config, profile=self.profile)
        self.browser.test_driver()
        self.browser.go_to(self.config.automsr.rewards_homepage)

    def end_session(self) -> None:
        """
        Terminate the session.
        """

        self.browser.driver.quit()

    def get_dashboard(self) -> Dashboard:
        """
        Retrieve a Dashboard object from the current page.

        This method is expected to be run inside the Rewards homepage.
        """

        raw_data: Dict[str, Any] = self.browser.execute_script("return dashboard;")
        dashboard = Dashboard(**raw_data)
        return dashboard

    def execute_pc_searches(self, dashboard: Dashboard) -> None:
        """
        Execute PC searches, if needed.
        """

        if (amount := dashboard.amount_of_pc_searches()) == 0:
            logger.info("No PC search is needed")
            return
        else:
            logger.info("Starting PC searches")

        assert amount > 0
        return self._execute_searches(
            amount=amount, user_agent=self.browser.user_agents.desktop
        )

    def execute_mobile_searches(self, dashboard: Dashboard) -> None:
        """
        Execute Mobile searches, if needed.
        """

        if (amount := dashboard.amount_of_mobile_searches()) == 0:
            logger.info("No Mobile search is needed")
            return
        else:
            logger.info("Starting Mobile searches")

        assert amount > 0
        return self._execute_searches(
            amount=amount, user_agent=self.browser.user_agents.mobile
        )

    def _execute_searches(
        self, amount: int = 1, user_agent: Optional[str] = None
    ) -> None:
        """
        Helper method to execute `amount` searches.

        Can specify a custom `user_agent` to use.
        """

        assert amount >= 1, "Invalid value!"

        safe_amount = amount + 5
        logger.debug("Original amount of searches: %s", amount)
        logger.debug("Safe amount of searches: %s", safe_amount)

        old_user_agent: Optional[str] = None
        if user_agent is not None:
            logger.debug("Changing user-agent to: %s", user_agent)
            old_user_agent = self.browser.get_user_agent()
            self.browser.change_user_agent(user_agent=user_agent)

        search_generator = RandomSearchGenerator()
        sleep_time = search_generator.sleep_time()
        query = search_generator.query_gen()

        self.browser.go_to_bing()

        for i in tqdm(range(safe_amount)):
            logger.debug("Executing search: %s/%s", i + 1, safe_amount)

            # we retrieve the element in the for-loop since the page is reloaded, thus the element can be invalidated
            element: WebElement = self.browser.driver.find_element(
                by=By.ID, value="sb_form_q"
            )

            # send the next item of the generator to the input field
            element.send_keys(next(query))

            # sleep to prevent issues with Selenium interacting with the page
            time.sleep(0.5)

            # send ENTER to perform a search
            element.send_keys(Keys.ENTER)

            # sleep a certain amount of time
            time.sleep(sleep_time)

        if user_agent is not None:
            logger.debug("Restoring original user-agent: %s", old_user_agent)
            assert old_user_agent is not None
            self.browser.change_user_agent(user_agent=old_user_agent)

    def execute_promotions(self, dashboard: Dashboard) -> None:
        """
        Execute all completable promotions.
        """

        # handle promotional items differently, since there is another way
        # to retrieve the corresponding WebElement
        promotional_item = dashboard.get_promotional_item()
        if promotional_item is not None:
            try:
                self.browser.open_promotion(
                    promotion=promotional_item, element_id="promo-item"
                )
            except Exception as e:
                logger.error("Exception caught: %s", e)
                logger.warning("Promotion '%s' will be skipped", promotional_item)
            finally:
                # simulate navigation
                logger.debug("Re-opening Rewards homepage to simulate navigation.")
                self.browser.go_to_rewards()
                logger.debug("Sleeping to simulate a real user behavior.")
                time.sleep(2)

        # handle all other promotions
        promotions: List[Promotion] = dashboard.get_completable_promotions()

        for promotion in promotions:
            logger.info("Executing promotion: %s", promotion.title)
            try:
                self._execute_promotion(promotion=promotion)
            except Exception as e:
                logger.error("Exception caught: %s", e)
                logger.warning("Promotion '%s' will be skipped", promotion)
            finally:
                # simulate navigation
                logger.debug("Re-opening Rewards homepage to simulate navigation.")
                self.browser.go_to_rewards()
                logger.debug("Sleeping to simulate a real user behavior.")
                time.sleep(2)

    def _execute_promotion(self, promotion: Promotion) -> None:
        """
        Execute a specific promotion.
        """

        driver = self.browser.driver
        self.browser.open_promotion(promotion=promotion)

        # Sleep to simulate user behavior and to let JS load the page
        time.sleep(3)

        # Determine which promotion is currently on
        if promotion.promotionType == PromotionType.QUIZ:
            # Try to determine which quiz we are dealing with
            try:
                _ = driver.find_element(by=By.ID, value="btoption0")
            except NoSuchElementException:
                pass
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.CHOICE_BETWEEN_TWO
                )

            # If the quiz is not a choice-between-two, try with the three-questions-N-answers
            try:
                start_button = driver.find_element(by=By.ID, value="rqStartQuiz")
                # if we find the button, we click it and then wait a little bit for JS
                # to load the answers in the DOM
                start_button.click()
                time.sleep(2)
            except NoSuchElementException as e:
                # if we are here, the quiz can be either already started,
                # or something's broken; in the latter case, we raise an error
                try:
                    driver.find_element(by=By.ID, value="currentQuestionContainer")
                except NoSuchElementException:
                    logger.error(
                        "Cannot find neither the start button, nor the quiz container!"
                    )
                    raise CannotFindStartButtonException() from e
                else:
                    logger.info("Quiz already started.")

            # Check if the quiz is 8-answers
            try:
                _ = driver.find_element(by=By.ID, value="rqAnswerOption7")
            except NoSuchElementException:
                pass
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.THREE_QUESTIONS_EIGHT_ANSWERS
                )

            # Check if the quiz is 4-answers
            try:
                _ = driver.find_element(by=By.ID, value="rqAnswerOption3")
            except NoSuchElementException as e:
                raise CannotDetermineQuizTypeException() from e
            else:
                return self._execute_quiz_promotion(
                    quiz_type=QuizType.THREE_QUESTIONS_FOUR_ANSWERS
                )

        elif promotion.promotionType in (
            PromotionType.URL_REWARD,
            PromotionType.WELCOME_TOUR,
        ):
            logger.info("Promotion executed by just opening the destination url.")
        else:
            raise ValueError(
                f"Cannot execute promotion with type: {promotion.promotionType}"
            )

    def _execute_quiz_promotion(self, quiz_type: QuizType) -> None:
        """
        Execute a specific promotion that is a quiz with a known type.

        The method assumes that, if a quiz type requires for a Start button to be clicked,
        this is done by the caller.
        """

        driver = self.browser.driver

        logger.info("Resolving promotion as quiz with type: %s", quiz_type)

        if quiz_type is QuizType.CHOICE_BETWEEN_TWO:
            answer = driver.find_element(by=By.ID, value="btoption0")
            answer.click()
        elif quiz_type in (
            QuizType.THREE_QUESTIONS_FOUR_ANSWERS,
            QuizType.THREE_QUESTIONS_EIGHT_ANSWERS,
        ):
            # safety breaks
            loop_break = 100
            loop_counter = 1

            # Create the `max_answer_index` based on the provided quiz type.
            if quiz_type is QuizType.THREE_QUESTIONS_FOUR_ANSWERS:
                max_answer_index = 4
            elif quiz_type is QuizType.THREE_QUESTIONS_EIGHT_ANSWERS:
                max_answer_index = 8
            else:
                raise ValueError(f"Invalid quiz type provided: {quiz_type}")

            current_answer_index = 0

            while loop_counter < loop_break:
                # check if quiz is finished
                try:
                    driver.find_element(by=By.ID, value="quizCompleteContainer")
                except NoSuchElementException:
                    logger.debug("Quiz still in progress.")
                else:
                    logger.info("Quiz finished.")
                    return

                # if not, continue with clicking answers
                answer_id = f"rqAnswerOption{current_answer_index}"
                answer_element = driver.find_element(by=By.ID, value=answer_id)
                answer_element.click()

                # update the answer index
                current_answer_index = (current_answer_index + 1) % max_answer_index

                # update the loop counter
                loop_counter += 1

            # if we are here, we didn't finish the quiz
            logger.warning("Quiz not finished!")
            return
        else:
            raise ValueError(f"Quiz type not supported: {quiz_type}")


@define
class MultipleTargetsExecutor:
    """
    Executor class driving the completion of all the tasks associated with Rewards.

    This executor will target multiple profiles automatically.
    """

    config: Config

    def execute(self) -> None:
        """
        Spawn a SingleTargetExecutor for every profile specified in the `config` file.
        """

        for profile in self.config.automsr.profiles:
            logger.info("Profile under execution: %s", profile)
            executor = SingleTargetExecutor(config=self.config, profile=profile)
            executor.execute()



import json
import logging
import os
import sys
from enum import Enum, auto
from pathlib import Path
from typing import Dict, List, Optional

from attrs import define

logger = logging.getLogger(__name__)

ENV_HOME = os.environ.get("HOME")
ENV_LOCALAPPDATA = os.environ.get("LOCALAPPDATA")


class ChromeVariant(Enum):
    """
    Chrome variants available
    """

    CHROME = auto()
    CHROME_CANARY = auto()
    CHROMIUM = auto()


class OutputFormat(Enum):
    """
    Types of output formats available for printing found Chrome profiles o stdout.
    """

    LIST = "list"
    JSON = "json"
    PRETTY_JSON = "pretty-json"


def get_platform() -> str:
    """
    Returns a simplified version of the system platform in use.
    """

    retval = "linux"  # assumes linux env if not win32 or darwin

    if sys.platform.startswith("win32"):
        retval = "windows"
    elif sys.platform.startswith("darwin"):
        retval = "macOS"

    return retval


@define
class ChromeProfile:
    displayed_name: str
    path: Path

    RESERVED_NAMES = ["System Profile", "Guest Profile"]

    def as_dict(self) -> Dict[str, str]:
        return dict(displayed_name=self.displayed_name, path=str(self.path.resolve()))

    @classmethod
    def from_directory(cls, path: Path) -> Optional["ChromeProfile"]:
        """
        Returns a Chrome Profile based on a directory.

        Returns `None` if the provided `path` is not a valid Chrome Profile.
        """

        # If the provided path is not an existing directory, returns None
        if not path.is_dir():
            return None

        # If the directory name is one of the reserved name for Google profiles
        # not matching any user profile, returns None
        if path.name in cls.RESERVED_NAMES:
            return None

        # If this file doesn't exist, it's not a profile
        preferences_path = path / "Preferences"
        if not preferences_path.is_file():
            return None

        preferences_dict = json.load(open(preferences_path, encoding="utf-8"))
        displayed_name = preferences_dict.get("profile", {}).get("name", "")
        return cls(displayed_name=displayed_name, path=path)


@define
class ProfilesExecutor:
    chrome_variant: ChromeVariant = ChromeVariant.CHROME
    profiles_root_path: Optional[Path] = None

    CHROME_PROFILES_LOCATIONS = {
        "macOS": {
            ChromeVariant.CHROME: f"{ENV_HOME}/Library/Application Support/Google/Chrome",
            ChromeVariant.CHROME_CANARY: f"{ENV_HOME}/Library/Application Support/Google/Chrome",
            ChromeVariant.CHROMIUM: f"{ENV_HOME}/Library/Application Support/Google/Chrome",
        },
        "windows": {
            ChromeVariant.CHROME: f"{ENV_LOCALAPPDATA}\\Google\\Chrome\\User Data",
            ChromeVariant.CHROME_CANARY: f"{ENV_LOCALAPPDATA}\\Google\\Chrome SxS\\User Data",
            ChromeVariant.CHROMIUM: f"{ENV_HOME}/Library/Application Support/Chromium",
        },
        "linux": {
            ChromeVariant.CHROME: f"{ENV_HOME}/.config/google-chrome",
            ChromeVariant.CHROME_CANARY: f"{ENV_HOME}/.config/google-chrome-beta",
            ChromeVariant.CHROMIUM: f"{ENV_HOME}/.config/chromium",
        },
    }

    def get_profiles_root_path(self) -> Path:
        """
        Based on current state, get the root path of Chrome profiles.
        """

        if self.profiles_root_path is not None:
            logger.info(
                "Profiles root path manually set to: %s", self.profiles_root_path
            )
            return self.profiles_root_path

        platform = get_platform()
        logger.debug("Platform to use: %s", platform)
        logger.debug("Chrome variant to use: %s", self.chrome_variant)

        profiles_root_path_str = self.CHROME_PROFILES_LOCATIONS[platform][
            self.chrome_variant
        ]
        profiles_root_path = Path(profiles_root_path_str)
        logger.info(
            "Profiles root path automatically found to be: %s", profiles_root_path
        )
        return profiles_root_path

    def get_profiles(self) -> List[ChromeProfile]:
        """
        Return the list of Chrome profiles found.
        """

        root_path = self.get_profiles_root_path()

        # Safety check
        if not root_path.exists():
            raise FileNotFoundError(root_path)

        chrome_profiles: List[ChromeProfile] = []

        for directory in root_path.iterdir():
            chrome_profile = ChromeProfile.from_directory(path=directory)
            if chrome_profile is None:
                continue

            logger.info("Found Chrome profile: %s", chrome_profile)
            chrome_profiles.append(chrome_profile)

        return chrome_profiles

    def print_profiles(self, output_format: OutputFormat = OutputFormat.LIST) -> None:
        """
        Get the list of Chrome profiles found, then print it to stdout.

        It's possible to specify a type for printing the values obtained to stdout.
        """

        profiles: List[ChromeProfile] = self.get_profiles()

        if output_format is OutputFormat.LIST:
            data = "\n".join([str(profile) for profile in profiles])

        elif output_format is OutputFormat.JSON:
            data = json.dumps([profile.as_dict() for profile in profiles])

        elif output_format is OutputFormat.PRETTY_JSON:
            data = json.dumps([profile.as_dict() for profile in profiles], indent=4)

        else:
            raise ValueError(output_format)

        data += "\n"  # add manually a final newline
        sys.stdout.write(data)



import datetime
import logging
import smtplib
from email.message import EmailMessage
from enum import Enum
from typing import Optional

from attr import define, field

from automsr.config import Config

logger = logging.getLogger(__name__)
NO_SENDER_ERR = "No sender email found!"
NO_SENDER_PSW_ERR = "No sender password found!"
NO_RECIPIENT_ERR = "No recipient specified!"
NO_HOST_ERR = "No SMTP host specified!"
NO_PORT_ERR = "No port specified!"
HOST_NOT_REACHABLE = "Cannot reach host! Check internet connection or hostname"
SHOULD_NOT_CREATE_CONN = (
    "Cannot create Email Connection if should not send any email! Check config"
)


class RewardsStatusEnum(Enum):
    FAILURE = "FAILURE"
    SUCCESS = "SUCCESS"


@define
class RewardsStatus:
    email: str
    status: RewardsStatusEnum
    message: str = ""

    def _set_status_and_message(self, status: RewardsStatusEnum, message: str):
        self.status = status
        self.message = message

    def set_success(self, message: str = ""):
        self._set_status_and_message(RewardsStatusEnum.SUCCESS, message)

    def set_failure(self, message: str = ""):
        self._set_status_and_message(RewardsStatusEnum.FAILURE, message)

    def to_plain(self):
        msg = f"{self.email} - {self.status.value}"
        if self.message:
            msg += f" - {self.message}"
        return msg

    def to_html(self):
        color = "red" if self.status is RewardsStatusEnum.FAILURE else "green"
        msg = f'<p><b>{self.email} - <font color="{color}">{self.status.value}</font></b></p>'
        if self.message:
            msg += f"<p><b>Message:</b> {self.message}</p>"
        return msg


class RewardsEmailMessage(EmailMessage):
    """An email message customized for Auto MSR"""

    now = datetime.date.today()
    prefix = "[AUTOMSR]"

    success_subject = f"{prefix} {now} SUCCESS"
    failure_subject = f"{prefix} {now} FAILURE"

    @classmethod
    def get_message(
        cls,
        from_email: str,
        to_email: str,
        subject: str,
        content: str = "",
        content_html: str = "",
    ):
        msg = cls()

        # setup message
        msg["From"] = from_email
        msg["To"] = to_email

        msg["Subject"] = subject

        msg.set_content(content)
        if content_html:
            msg.add_alternative(content_html, subtype="html")

        return msg

    @classmethod
    def get_success_message(
        cls, from_email: str, to_email: str, content: str = "", content_html: str = ""
    ):
        return cls.get_message(
            from_email=from_email,
            to_email=to_email,
            subject=cls.success_subject,
            content=content,
            content_html=content_html,
        )

    @classmethod
    def get_failure_message(
        cls, from_email: str, to_email: str, content: str = "", content_html: str = ""
    ):
        return cls.get_message(
            from_email=from_email,
            to_email=to_email,
            subject=cls.failure_subject,
            content=content,
            content_html=content_html,
        )


@define(slots=False)
class EmailConnection:
    sender: str
    password: str
    recipient: str
    host: str
    port: int = field(converter=int)
    tls: bool = field(default=False, converter=bool)
    smtp: smtplib.SMTP = field(init=False)

    def open(self) -> None:
        """
        Open an SMTP connection and try to log in with mail server.
        """

        self.smtp.ehlo()
        if self.tls:
            self.smtp.starttls()
        self.smtp.login(self.sender, self.password)

    def close(self) -> None:
        """
        Closes an SMTP established connection.
        """

        try:
            self.smtp.quit()
        except smtplib.SMTPServerDisconnected:
            pass
        finally:
            logger.debug("SMTP connection closed")

    def test_connection(self) -> None:
        """
        Test if the provided connection is successful.
        Any exception is raised to the caller.
        """

        self.open()
        self.close()
        logger.info("Email connection was successful")

    def send_success_message(self):
        msg = RewardsEmailMessage.get_success_message(self.sender, self.recipient)
        self.smtp.send_message(msg)

    def send_failure_message(self):
        msg = RewardsEmailMessage.get_failure_message(self.sender, self.recipient)
        self.smtp.send_message(msg)

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@define(slots=False)
class OutlookEmailConnection(EmailConnection):
    host: str = "smtp-mail.outlook.com"
    port: int = 587
    tls: bool = True


@define(slots=False)
class GmailEmailConnection(EmailConnection):
    host: str = "smtp.gmail.com"
    port: int = 587
    tls: bool = True


@define
class EmailConnectionFactory:
    """
    Connection factory; will try to create automagically a connection based on the sender address.
    """

    config: Config

    def get_connection(self) -> Optional[EmailConnection]:
        """
        Get an email connection based on the config.

        >>> from pathlib import Path
        >>> config_path = Path("tests/configs/config.example.yaml")
        >>> _config = Config.from_yaml(config_path)
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().__class__
        <class 'automsr.mail.GmailEmailConnection'>

        >>> _config.email.sender = "foo@outlook.com"
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().__class__
        <class 'automsr.mail.OutlookEmailConnection'>

        >>> _config.email.sender = "foo@foobar.com"
        >>> factory = EmailConnectionFactory(config=_config)
        >>> factory.get_connection().__class__
        <class 'automsr.mail.EmailConnection'>
        """

        config = self.config.email

        if not config.enable:
            logger.info("Email send was disabled by config")
            return None

        sender = config.sender
        assert sender is not None
        domain = sender.split("@")[1]
        logger.debug("Sender domain: %s", domain)

        assert config.sender_password is not None
        assert config.recipient is not None

        # Handle Gmail emails
        if domain == "gmail.com":
            logger.debug("Gmail domain found")
            return GmailEmailConnection(
                sender=sender,
                password=config.sender_password.get_secret_value(),
                recipient=config.recipient,
            )

        # Handle Outlook emails
        elif domain.split(".")[0] in ("live", "outlook", "hotmail"):
            logger.debug("Outlook domain found")
            return OutlookEmailConnection(
                sender=sender,
                password=config.sender_password.get_secret_value(),
                recipient=config.recipient,
            )

        # Handle custom domains
        else:
            logger.debug("Custom domain found")
            assert config.host is not None
            assert config.port is not None
            assert config.tls is not None

            return EmailConnection(
                sender=sender,
                password=config.sender_password.get_secret_value(),
                recipient=config.recipient,
                host=config.host,
                port=config.port,
                tls=config.tls,
            )



import logging
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Callable

from attr import define, field

from automsr.browser.profile import OutputFormat, ProfilesExecutor
from automsr.config import Config
from automsr.executor import MultipleTargetsExecutor

logger = logging.getLogger(__name__)

DEFAULT_CONFIG_PATH = Path("config.yaml")


@define
class Args:
    """
    Class mapping of CLI arguments.
    """

    # Method to invoke with a specific subparser
    func: Callable[["Args"], Any]

    # Common args
    config: Path = DEFAULT_CONFIG_PATH
    verbose: bool = False

    # Run args

    # Profiles args
    format: OutputFormat = field(default=OutputFormat.LIST, converter=OutputFormat)


def run(args: Args) -> None:
    """
    Method to invoke when `run` is executed.
    """

    config = Config.from_yaml(args.config)
    executor = MultipleTargetsExecutor(config=config)
    executor.execute()
    logger.info("Execution finished!")


def profiles(args: Args) -> None:
    """
    Method to invoke when `profiles` is executed.
    """

    config = Config.from_yaml(args.config)
    executor = ProfilesExecutor(profiles_root_path=config.selenium.profiles_root)
    executor.print_profiles(output_format=args.format)


def add_common_flags(parser: ArgumentParser) -> None:
    """
    Add common flags to a generic parser.

    Flags provided:
    * -c, --config
    * -v, --verbose
    """

    parser.add_argument(
        "-c",
        "--config",
        type=Path,
        default=DEFAULT_CONFIG_PATH,
        help=f"Path to config file. Defaults to {DEFAULT_CONFIG_PATH!s}",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Increase verbosity"
    )


def add_run_flags(parser: ArgumentParser) -> None:
    """
    Add `run` flags to a generic parser.

    Flags provided: no one.
    """

    parser.set_defaults(func=run)


def add_profiles_flags(parser: ArgumentParser) -> None:
    """
    Add `profiles` flags to a generic parser.

    Flags provided:
    * -f, --format
    """

    parser.add_argument(
        "-f",
        "--format",
        default=OutputFormat.LIST.value,
        choices=[v.value for v in list(OutputFormat)],
        help=f"Output format for profiles. Defaults to {OutputFormat.LIST.value}",
    )
    parser.set_defaults(func=profiles)


def cli() -> None:
    # Construct the base parser
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(required=True, help="Available subcommands.")

    # Construct the `run` parser
    run_parser = subparsers.add_parser(
        name="run",
        help="Run AutoMSR on profiles specified in the config file, then exit.",
    )
    add_common_flags(parser=run_parser)
    add_run_flags(parser=run_parser)

    # Construct the `profiles` parser
    profiles_parser = subparsers.add_parser(
        name="profiles",
        help="Retrieve the Chrome profiles found in the local machine with their full path.",
    )
    add_common_flags(parser=profiles_parser)
    add_profiles_flags(parser=profiles_parser)

    # Parse arguments
    try:
        raw_args = vars(parser.parse_args())
    except TypeError:
        parser.print_help()
        sys.exit(1)
    args = Args(**raw_args)

    # Handle verbosity
    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level)

    # Set selenium and urllib3 logging level to WARNING,
    # regardless of the tool's selected logging level
    logging.getLogger("selenium").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    # Execute subparser functionality
    args.func(args)


if __name__ == "__main__":
    cli()



import logging
from pathlib import Path
from typing import Any, Optional

from attr import define
from selenium.common import WebDriverException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.chrome.webdriver import WebDriver as ChromeWebDriver
from selenium.webdriver.common.by import By

from automsr.config import Config, Defaults, Profile
from automsr.datatypes.dashboard import Promotion

logger = logging.getLogger(__name__)


class BrowserException(Exception):
    """
    Base class for Browser exceptions.
    """


class CannotChangeUserAgentException(BrowserException):
    """
    Exception raised when the browser is unable to change User Agent.
    """


class CannotStartBrowserException(BrowserException):
    """
    Exception raised when Selenium is unable to start a new Browser Session.
    """


@define
class BrowserOptions:
    """
    Abstraction of Browser Options, and wrapper around Selenium Chrome Options.
    """

    profiles_root: Path
    profile_directory_name: str

    @classmethod
    def from_config(cls, config: Config, profile: Profile) -> "BrowserOptions":
        """
        Generate the Options based on the profiles_root specified in the `config`, and a chosen `profile`.
        """

        profiles_root = config.selenium.profiles_root
        profile_directory_name = profile.profile

        # safety check
        profile_path = profiles_root / profile_directory_name
        if not profile_path.is_dir():
            raise FileNotFoundError(
                f"Profile `{profile_directory_name}` not found in directory: {profiles_root}"
            )

        return cls(
            profiles_root=profiles_root, profile_directory_name=profile_directory_name
        )

    def as_chromium(self) -> Options:
        """
        Returns the object as Chromium Options.

        Exhaustive list of Chromium switches: https://peter.sh/experiments/chromium-command-line-switches/
        """

        options = Options()
        options.add_argument("--start-maximized")
        options.add_argument(f"--user-data-dir={self.profiles_root!s}")
        options.add_argument(f"--profile-directory={self.profile_directory_name}")
        return options


@define
class UserAgent:
    desktop: str = Defaults.desktop_useragent
    mobile: str = Defaults.mobile_useragent


@define
class RewardsUrl:
    bing: str = Defaults.bing_homepage
    rewards: str = Defaults.rewards_homepage


@define
class Browser:
    driver: ChromeWebDriver

    user_agents: UserAgent = UserAgent()
    urls: RewardsUrl = RewardsUrl()

    def change_user_agent(self, user_agent: str, strict: bool = True) -> None:
        """
        Change user agent in the current driver.

        If `check` is True, then the execution will fail if the user agent is not changed.

        Available commands:
        Driver side:
            https://github.com/SeleniumHQ/selenium/blob/selenium-4.11.2-python/py/selenium/webdriver/chromium/remote_connection.py#L36
        Server side:
            https://github.com/SeleniumHQ/selenium/blob/selenium-4.11.2-python/dotnet/src/webdriver/DevTools/Network.cs#L79
        """

        driver_command = "executeCdpCommand"

        server_command = "Network.setUserAgentOverride"
        server_command_args = dict(userAgent=user_agent)

        logger.debug(f"Trying to change user-agent to: {user_agent}")

        self.driver.execute(
            driver_command, {"cmd": server_command, "params": server_command_args}
        )

        actual_user_agent = self.get_user_agent()
        if actual_user_agent != user_agent:
            if strict:
                raise CannotChangeUserAgentException("Cannot set a new user-agent!")
            else:
                logger.warning(
                    f"Cannot set a new user-agent! Current user-agent: {actual_user_agent}"
                )
        else:
            logger.debug(f"Changed user-agent to: {actual_user_agent}")

    @classmethod
    def from_config(cls, config: Config, profile: Profile) -> "Browser":
        """
        Construct a Browser from a `config` and a `profile` provided as inputs.
        """

        if (path := config.selenium.chromedriver_path) is not None:
            chromedriver_path = str(path)
        else:
            chromedriver_path = None
        logger.debug("Chromedriver path: %s", chromedriver_path)

        user_agents = UserAgent(
            desktop=config.automsr.desktop_useragent,
            mobile=config.automsr.mobile_useragent,
        )
        urls = RewardsUrl(
            bing=config.automsr.bing_homepage,
            rewards=config.automsr.rewards_homepage,
        )

        logger.debug("User agents used: %s", user_agents)
        logger.debug("Rewards URLs used: %s", urls)

        options = BrowserOptions.from_config(config=config, profile=profile)
        logger.debug("Browser options: %s", options)

        chromium_options = options.as_chromium()
        service = Service(chromedriver_path=chromedriver_path)
        try:
            driver = ChromeWebDriver(options=chromium_options, service=service)
            driver.implicitly_wait(time_to_wait=3)
        except WebDriverException as e:
            exception = CannotStartBrowserException(
                "Cannot create a new Chrome Session! Maybe there is already one process running?"
            )
            raise exception from e

        return cls(driver=driver, user_agents=user_agents, urls=urls)

    def test_driver(self) -> None:
        """
        Test if the driver is working correctly.
        """

        test_url = "https://www.google.com"
        self.driver.get(test_url)

    def go_to(self, url: str) -> None:
        """
        Change page following the provided url.
        """

        self.driver.get(url)

    def execute_script(self, script: str) -> Any:
        """
        Execute one or more JS instructions and returns the result.
        """

        return self.driver.execute_script(script=script)

    def go_to_bing(self) -> None:
        """
        Change page to Bing homepage.
        """

        return self.go_to(self.urls.bing)

    def go_to_rewards(self) -> None:
        """
        Change page to Rewards homepage.
        """

        return self.go_to(self.urls.rewards)

    def get_user_agent(self) -> str:
        """
        Returns the current User-Agent.
        """

        return str(self.driver.execute_script("return navigator.userAgent;"))

    def change_window(self, handle_name: str) -> None:
        """
        Switch window to the one specified by the handle name.
        """

        windows = self.driver.window_handles
        window_index = windows.index(handle_name)
        window = windows[window_index]
        self.driver.switch_to.window(window)

    def open_promotion(
        self, promotion: Promotion, element_id: Optional[str] = None
    ) -> None:
        """
        Open correctly a Promotion found in the Rewards homepage.

        If the `element-id` is provided, use that to retrieve the WebElement from the page.
        """

        current_url: str = self.driver.current_url
        rewards_url: str = self.urls.rewards

        if not current_url.startswith(rewards_url):
            logger.debug("Changing current page to Rewards homepage.")
            self.go_to_rewards()

        if element_id is not None:
            logger.debug("Searching for promotion with id: %s", element_id)
            if not element_id:
                raise ValueError("Empty element id!")
            promotion_element = self.driver.find_element(by=By.ID, value=element_id)
        else:
            css_selector_value = (
                f'.rewards-card-container[data-bi-id="{promotion.name}"]'
            )
            logger.debug(
                "Searching for promotion with css selector: %s", css_selector_value
            )
            promotion_element = self.driver.find_element(
                by=By.CSS_SELECTOR, value=css_selector_value
            )

        logger.debug("Promotion found! Clicking it to trigger the promotion start.")

        # store window handles before the click
        handles_before = self.driver.window_handles

        # click the element
        promotion_element.click()

        # store window handles after the click
        handles_after = self.driver.window_handles

        handle = set(handles_after).difference(handles_before).pop()
        self.change_window(handle_name=handle)
